#  Part of the LLVM Project, under the Apache License v2.0 with LLVM Exceptions.
#  See https://llvm.org/LICENSE.txt for license information.
#  SPDX-License-Identifier: Apache-2.0 WITH LLVM-exception

from .._memref_transform_ops_gen import *
from .._memref_transform_ops_gen import _Dialect

try:
    from ...ir import *
    from ...dialects import transform
    from .._ods_common import _cext as _ods_cext
except ImportError as e:
    raise RuntimeError("Error loading imports from extension module") from e

from typing import Optional, overload, Union


@_ods_cext.register_operation(_Dialect, replace=True)
class MemRefAllocaToGlobalOp(MemRefAllocaToGlobalOp):
    """Specialization for MemRefAllocaToGlobalOp class."""

    @overload
    def __init__(
        self,
        get_global_type: Type,
        global_type: Type,
        alloca: Union[Operation, OpView, Value],
        *,
        loc=None,
        ip=None,
    ):
        ...

    @overload
    def __init__(self, alloca: Union[Operation, OpView, Value], *, loc=None, ip=None):
        ...

    def __init__(
        self,
        get_global_type_or_alloca: Union[Operation, OpView, Type, Value],
        global_type_or_none: Optional[Type] = None,
        alloca_or_none: Optional[Union[Operation, OpView, Value]] = None,
        *,
        loc=None,
        ip=None,
    ):
        if isinstance(get_global_type_or_alloca, Type):
            get_global_type = get_global_type_or_alloca
            global_type = global_type_or_none
            alloca = alloca_or_none
        else:
            get_global_type = transform.AnyOpType.get()
            global_type = transform.AnyOpType.get()
            alloca = get_global_type_or_alloca

        super().__init__(
            get_global_type,
            global_type,
            alloca,
            loc=loc,
            ip=ip,
        )


@_ods_cext.register_operation(_Dialect, replace=True)
class MemRefMultiBufferOp(MemRefMultiBufferOp):
    """Specialization for MemRefMultiBufferOp class."""

    @overload
    def __init__(
        self,
        transformed_type: Type,
        target: Union[Operation, OpView, Value],
        factor: Union[int, IntegerAttr],
        *,
        skip_analysis: Optional[bool] = None,
        loc=None,
        ip=None,
    ):
        ...

    @overload
    def __init__(
        self,
        target: Union[Operation, OpView, Value],
        factor: Union[int, IntegerAttr],
        *,
        skip_analysis: Optional[bool] = None,
        loc=None,
        ip=None,
    ):
        ...

    def __init__(
        self,
        transformed_type_or_target: Type,
        target_or_factor: Union[int, IntegerAttr, Operation, OpView, Value] = None,
        factor_or_none: Optional[Union[int, IntegerAttr]] = None,
        *,
        skip_analysis: Optional[bool] = None,
        loc=None,
        ip=None,
    ):
        if isinstance(transformed_type_or_target, Type):
            transformed_type = transformed_type_or_target
            target = target_or_factor
            factor = factor_or_none
        else:
            transformed_type = transform.AnyOpType.get()
            target = transformed_type_or_target
            factor = target_or_factor

        super().__init__(
            transformed_type,
            target,
            factor,
            skip_analysis=skip_analysis,
            loc=loc,
            ip=ip,
        )



# SPDX-FileCopyrightText: Copyright (c) 2026 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: Apache-2.0

import operator
from numba_cuda_mlir import types
from numba_cuda_mlir.numba_cuda.typing.templates import (
    AbstractTemplate,
    Registry,
)

registry = Registry()

_COMPARISON_OPS = [
    operator.eq,
    operator.ne,
    operator.lt,
    operator.le,
    operator.gt,
    operator.ge,
]

for _op in _COMPARISON_OPS:

    class UnicodeComparisonTemplate(AbstractTemplate):
        key = _op
        metadata = {"target": "cuda"}

        def generic(self, args, kws):
            a, b = args
            accepted = (types.UnicodeType, types.StringLiteral)
            if isinstance(a, accepted) and isinstance(b, accepted):
                return types.boolean(a, b)

    registry.register(UnicodeComparisonTemplate)
    registry.register_global(_op, types.Function(UnicodeComparisonTemplate))



# SPDX-FileCopyrightText: Copyright (c) 2026 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
"""
Union support for numba_cuda_mlir - similar to ctypes structures
"""

from collections import namedtuple, Counter
from numba_cuda_mlir.numba_cuda import types
from numba_cuda_mlir.type_defs.aggregate_types import AggregateType, UnionType
from typing import Any
import ctypes
from .struct import StructTypeWrapper, StructInstance
from numba_cuda_mlir.lowering.union import _lower_union_construction_impl
from numba_cuda_mlir.descriptor import mlir_target
from numba_cuda_mlir.numba_cuda.typing.templates import ConcreteTemplate
from numba_cuda_mlir import typing

# Counter for generating unique union names if no name is provided
_anon_union_counter = 0


class UnionInstance:
    """
    Runtime instance of a union that can be passed to kernels.

    C-style union where all variants share the same memory location.
    Any variant can be read or written at any time.
    """

    def __init__(self, union_type: UnionType, **kwargs):
        self._union_type = union_type
        self._variants = {}

        # Initialize all variants with default values
        for variant_name, variant_type in union_type.variants:
            self._variants[variant_name] = self._get_default_value(variant_type)

        # If specific variant values were provided, set them
        for variant_name, value in kwargs.items():
            if variant_name in self._variants:
                self._variants[variant_name] = value

    def _get_default_value(self, variant_type: types.Type) -> Any:
        """Get a default value for a variant type."""
        if isinstance(variant_type, types.Integer):
            return 0
        elif isinstance(variant_type, types.Float):
            return 0.0
        elif isinstance(variant_type, types.Boolean):
            return False
        elif isinstance(variant_type, AggregateType):
            # Nested struct
            return StructInstance(variant_type)
        else:
            return None

    def __getattr__(self, name: str) -> Any:
        """Allow accessing variants via dot notation."""
        if name.startswith("_"):
            # Private attributes go through normal lookup
            return object.__getattribute__(self, name)

        if name in self._variants:
            return self._variants[name]

        raise AttributeError(f"Union '{self._union_type.name}' has no variant '{name}'")

    def __setattr__(self, name: str, value: Any):
        """Allow setting variants via dot notation."""
        if name.startswith("_"):
            # Private attributes go through normal setattr
            object.__setattr__(self, name, value)
        elif hasattr(self, "_variants") and name in self._variants:
            # When setting a union variant, we update that variant
            # In C, this would overwrite the shared memory
            self._variants[name] = value
        else:
            object.__setattr__(self, name, value)

    def __repr__(self) -> str:
        variant_strs = [f"{name}=..." for name, _ in self._union_type.variants]
        return f"{self._union_type.name}(union of: {', '.join(variant_strs)})"


class UnionTypeWrapper:
    """
    Wrapper around UnionType that adds a constructor.

    This allows users to call `MyUnion(variant1=val1)` to create instances.
    """

    def __init__(self, union_type: UnionType):
        self._type = union_type
        self.name = union_type.name
        self.variants = union_type.variants
        self._constructor_template = None

    def __call__(self, **kwargs) -> UnionInstance:
        """Create an instance of this union."""
        return UnionInstance(self._type, **kwargs)

    def __repr__(self) -> str:
        return f"UnionType({self.name})"

    def __getattr__(self, name):
        return getattr(self._type, name)


def union(variants: list[tuple[str, types.Type]], name: str | None = None) -> UnionTypeWrapper:
    """
    Create a union type for use in CUDA kernels.

    This creates a C-style union where all variants share the same memory location.
    Any variant can be read or written at any time.

    Args:
        variants: List of tuples specifying variant names and types.
        name: Optional name for the union type. If not provided, generates a unique anonymous name.

    Returns:
        A UnionTypeWrapper that can be called to create instances

    Example:
        >>> from numba_cuda_mlir import union,
        >>> from numba import types
        >>>
        >>> MyUnion = union(
        ...     "MyUnion",
        ...     [
        ...         ("as_int", types.uint32),
        ...         ("as_float", types.float32),
        ...     ],
        ... )
        >>>
        >>> # Create an instance
        >>> u = MyUnion(as_int=42)
        >>> print(u.as_int)
        42
    """
    # Generate unique name for anonymous unions to avoid registry collisions
    if name is None:
        global _anon_union_counter
        name = f"AnonymousUnion_{_anon_union_counter}"
        _anon_union_counter += 1

    # Validate that variant names are unique
    variant_names = [variant[0] for variant in variants]
    if len(set(variant_names)) != len(variant_names):
        duplicates = [name for name, count in Counter(variant_names).items() if count > 1]
        raise ValueError(
            f"Union '{name}' has duplicate variant names: {duplicates}. "
            f"All variant names must be unique within a union definition."
        )

    # Convert  namedtuples to (name, type) tuples
    # Unwrap StructTypeWrapper if present to get the underlying type
    variant_tuples = []
    for variant in variants:
        variant_type = variant[1]
        # Unwrap StructTypeWrapper to get the actual AggregateType
        if isinstance(variant_type, StructTypeWrapper):
            variant_type = variant_type._type
        variant_tuples.append((variant[0], variant_type))

    # Create the UnionType first
    union_type = UnionType(name, variant_tuples)

    # Get the union's storage size (maximum of all variants)
    # This will be used as the storage type for bitfield structs
    union_storage_bits = union_type.size_bits

    # Propagate storage size to bitfield struct variants
    # This matches C behavior where bitfields in a union struct pack into the union's storage
    for _variant_name, variant_type in variant_tuples:
        if isinstance(variant_type, AggregateType) and variant_type.is_bitfield_struct:
            # This is a bitfield struct - set its union storage size
            variant_type.union_storage_bits = union_storage_bits

    # Wrap it so it can be called to create instances
    wrapper = UnionTypeWrapper(union_type)

    # Register the constructor as callable in device code
    _register_union_constructor(wrapper, union_type)

    return wrapper


def _register_union_constructor(wrapper: UnionTypeWrapper, union_type: UnionType):
    """
    Register a union constructor so it can be called in device code.

    This allows `MyUnion()` to work inside kernels, where it will allocate
    the union and return it.
    """

    typingctx = mlir_target.typing_context
    targetctx = mlir_target.target_context
    sig = typing.signature(union_type)

    # Create a template that makes this wrapper callable
    class union_constructor_template(ConcreteTemplate):
        key = wrapper
        cases = [sig]

    wrapper._constructor_template = union_constructor_template

    typingctx.insert_user_function(wrapper, union_constructor_template)

    def lower_union_wrapper_call(builder, target, args, kwargs):
        """Lower calls to this specific union constructor."""
        return _lower_union_construction_impl(None, builder, target, args, kwargs)

    targetctx.insert_func_defn([(lower_union_wrapper_call, wrapper, ())])



# SPDX-FileCopyrightText: Copyright (c) 2025 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: BSD-2-Clause

import functools


class _LLVMLiteRemoved:
    """Tombstone standing in for the removed ``llvmlite.ir`` module.

    These CUDA data models are instantiated during MLIR type inference, but the
    llvmlite ``be_type`` they would build is discarded (the MLIR-typed models
    live in ``numba_cuda_mlir.models``). The eager type construction has been
    replaced by ``None``; any remaining use raises, since llvmlite is no longer
    a dependency.
    """

    def __getattr__(self, name):
        raise NotImplementedError(
            "llvmlite has been removed; llvmlite IR is not available on the "
            f"MLIR path (attempted to use llvmlite.ir.{name})"
        )


ir = _LLVMLiteRemoved()

from numba_cuda_mlir.numba_cuda.datamodel.registry import DataModelManager, register
from numba_cuda_mlir.numba_cuda.datamodel import PrimitiveModel
from numba_cuda_mlir.numba_cuda.datamodel.models import StructModel
from numba_cuda_mlir.numba_cuda.extending import core_models as models
from numba_cuda_mlir.numba_cuda import types
from numba_cuda_mlir.numba_cuda.types.ext_types import (
    Dim3,
    GridGroup,
    CUDADispatcher,
    Bfloat16,
)


cuda_data_manager = DataModelManager()

register_model = functools.partial(register, cuda_data_manager)


@register_model(Dim3)
class Dim3Model(StructModel):
    def __init__(self, dmm, fe_type):
        members = [("x", types.int32), ("y", types.int32), ("z", types.int32)]
        super().__init__(dmm, fe_type, members)


@register_model(GridGroup)
class GridGroupModel(models.PrimitiveModel):
    def __init__(self, dmm, fe_type):
        # llvmlite be_type discarded on the MLIR path
        super().__init__(dmm, fe_type, None)


@register_model(types.Float)
class FloatModel(models.PrimitiveModel):
    def __init__(self, dmm, fe_type):
        if fe_type not in (types.float16, types.float32, types.float64):
            raise NotImplementedError(fe_type)
        # llvmlite be_type discarded on the MLIR path
        super().__init__(dmm, fe_type, None)


register_model(CUDADispatcher)(models.OpaqueModel)


@register_model(Bfloat16)
class _model___nv_bfloat16(PrimitiveModel):
    def __init__(self, dmm, fe_type):
        # llvmlite be_type discarded on the MLIR path
        super().__init__(dmm, fe_type, None)



# SPDX-FileCopyrightText: Copyright (c) 2026 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
from functools import lru_cache
import sys
from typing import Callable

if sys.version_info >= (3, 12):
    from typing import override
else:
    from typing_extensions import override
from numba_cuda_mlir.numba_cuda.core.imputils import Registry
from numba_cuda_mlir.numba_cuda import types
import functools


def _decorate_getattr(impl, ty, attr):
    """
    The upstream Numba version of the getattr lowering registration
    does not use functools.wraps, but we would like to filter out the
    upstream numba lowerings and only apply our own.
    """
    real_impl = impl

    if attr is not None:

        @functools.wraps(real_impl)
        def res(context, builder, typ, value, attr):
            return real_impl(context, builder, typ, value)

    else:

        @functools.wraps(real_impl)
        def res(context, builder, typ, value, attr):
            return real_impl(context, builder, typ, value, attr)

    res.signature = (ty,)
    res.attr = attr
    return res


def _decorate_setattr(impl, ty, attr):
    """
    Similar to _decorate_getattr but for setattr lowerings.
    Preserves __module__ so our lowerings aren't filtered out.
    """
    real_impl = impl

    if attr is not None:

        @functools.wraps(real_impl)
        def res(context, builder, sig, args, attr):
            return real_impl(context, builder, sig, args)

    else:

        @functools.wraps(real_impl)
        def res(context, builder, sig, args, attr):
            return real_impl(context, builder, sig, args, attr)

    # Signature is (target_type, value_type) for setattr lookup
    # Use types.Any for value_type to match any value type
    res.signature = ty, types.Any
    res.attr = attr
    return res


class LoweringRegistry(Registry):
    """
    Registry for MLIR-based lowering implementations.

    Each lowering module should create its own instance of this registry:
        registry = LoweringRegistry()
        lower = registry.lower
    """

    def __init__(self):
        super().__init__(name="numba_cuda_mlir")

    def lower_getattr(self, ty, attr):
        """
        Decorate an implementation of __getattr__ for type *ty* and
        the attribute *attr*.

        The decorated implementation will have the signature
        (context, builder, typ, val).
        """

        def decorate(impl):
            return self._decorate_attr(impl, ty, attr, self.getattrs, _decorate_getattr)

        return decorate

    def lower_getattr_generic(self, ty):
        """
        Decorate a generic implementation of __getattr__ for type *ty*.

        This overrides Numba's Registry.lower_getattr_generic to use our
        custom _decorate_getattr which preserves __module__ so that our
        lowerings aren't filtered out by _filter_numba_lowerings.

        The decorated implementation will have the signature
        (context, builder, typ, val, attr).
        """

        def decorate(impl):
            # Use attr=None to indicate generic getattr (handles any attribute)
            return self._decorate_attr(impl, ty, None, self.getattrs, _decorate_getattr)

        return decorate

    def lower_setattr_generic(self, ty):
        """
        Decorate a generic implementation of __setattr__ for type *ty*.

        This overrides Numba's Registry.lower_setattr_generic to use our
        custom _decorate_setattr which preserves __module__ so that our
        lowerings aren't filtered out.

        The decorated implementation will have the signature
        (context, builder, sig, args, attr).
        """

        def decorate(impl):
            # Use attr=None to indicate generic setattr (handles any attribute)
            return self._decorate_attr(impl, ty, None, self.setattrs, _decorate_setattr)

        return decorate



# SPDX-FileCopyrightText: Copyright (c) 2026 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: Apache-2.0

"""
NRT (Numba Runtime) constants and utilities.

NRT device functions are emitted as MLIR LLVM dialect ops by nrt_mlir.py.
The old NVRTC compilation path (nrt.cu) has been removed.
"""

from pathlib import Path

_NRT_DIR = Path(__file__).parent


def get_include():
    """Return the include path for the NRT headers."""
    return str(_NRT_DIR)


NRT_FUNCTIONS = frozenset(
    [
        "NRT_Allocate",
        "NRT_MemInfo_alloc",
        "NRT_MemInfo_init",
        "NRT_MemInfo_new",
        "NRT_Free",
        "NRT_dealloc",
        "NRT_MemInfo_destroy",
        "NRT_MemInfo_call_dtor",
        "NRT_MemInfo_data_fast",
        "NRT_MemInfo_alloc_aligned",
        "NRT_Allocate_External",
        "NRT_decref",
        "NRT_incref",
    ]
)


def needs_nrt_linking(asm: str) -> bool:
    """Check if the given assembly/PTX references NRT functions."""
    return any(fn in asm for fn in NRT_FUNCTIONS)



from pathlib import Path
import os
import re
from functools import lru_cache
from typing import overload
from numba_cuda_mlir.numba_cuda import itanium_mangler
from numba_cuda_mlir.numba_cuda import types


def parse_compute_capability(compute_cap: str) -> tuple[int, int]:
    match = re.match(r"sm_([1-9]\d?)(\d)(a)?", compute_cap)
    if match:
        return int(match.group(1)), int(match.group(2))
    else:
        raise ValueError(f"Invalid compute capability: {compute_cap}")


def format_arch(cc: tuple[int, int]) -> str:
    """
    Format a compute capability tuple as an sm_XX[a] string.

    Args:
        cc: Compute capability as (major, minor) tuple
    """
    return f"sm_{cc[0]}{cc[1]}"


def resolve_gpu_target(targetoptions: dict | None = None) -> dict[str, object]:
    if targetoptions is None:
        targetoptions = {}

    chip = targetoptions.get("chip")
    if chip:
        arch = chip
        cc = parse_compute_capability(arch)
        arch_suffix = arch.removeprefix(f"sm_{cc[0]}{cc[1]}")
        arch_specific_cc = (*cc, arch_suffix) if arch_suffix in ("a", "f") else cc
    else:
        cc = get_gpu_compute_capability(tuple)
        arch = format_arch(cc)
        arch_specific_cc = cc

    host_cc = get_gpu_compute_capability(tuple)
    host_arch = format_arch(host_cc)
    if cc < host_cc:
        linker_cc = host_cc
        linker_arch = host_arch
    else:
        linker_cc = arch_specific_cc
        linker_arch = arch

    return {
        "chip": arch,
        "cc": cc,
        "arch_specific_cc": arch_specific_cc,
        "host_cc": host_cc,
        "host_arch": host_arch,
        "linker_cc": linker_cc,
        "linker_arch": linker_arch,
    }


def resolve_target_options(targetoptions: dict[str, object]) -> dict[str, object]:
    target = resolve_gpu_target(targetoptions)
    targetoptions["chip"] = target["chip"]
    return targetoptions


@lru_cache(maxsize=1)
def get_cuda_toolkit_path() -> str | None:
    """Get CUDA toolkit root path from env vars, numba-cuda discovery, or system."""
    for var in ("CUDA_HOME", "CUDA_PATH"):
        val = os.environ.get(var)
        if val and os.path.isdir(os.path.join(val, "bin")):
            return val

    from numba_cuda_mlir.numba_cuda.cuda_paths import get_cuda_paths

    libdevice = get_cuda_paths().get("libdevice")
    if libdevice and libdevice.info:
        # Toolkit root is 3 levels up: toolkit/nvvm/libdevice/libdevice.10.bc
        root = os.path.dirname(os.path.dirname(os.path.dirname(libdevice.info)))
        if os.path.isdir(os.path.join(root, "bin")):
            return root

    if os.path.isdir("/usr/local/cuda/bin"):
        return "/usr/local/cuda"

    return None


@lru_cache(maxsize=1)
def get_cuda_runtime_version() -> tuple[int, int]:
    """
    Get the CUDA toolkit version as a (major, minor) tuple.

    Use ``nvrtcVersion()`` from libnvrtc, which reflects the version of
    installed CUDA toolkit.
    """
    from cuda.bindings import nvrtc

    err, major, minor = nvrtc.nvrtcVersion()
    if err != nvrtc.nvrtcResult.NVRTC_SUCCESS:
        raise RuntimeError(f"nvrtcVersion() failed: {err}")
    return (major, minor)


# CTK (major, minor) -> max PTX ISA version supported by that toolkit's libnvvm.
# Expressed as the integer used in the `+ptxNN` target feature string.
_CTK_TO_MAX_PTX: dict[tuple[int, int], int] = {
    (12, 8): 87,  # PTX 8.7
    (12, 9): 88,  # PTX 8.8
    (13, 0): 90,  # PTX 9.0
    (13, 1): 91,  # PTX 9.1
    (13, 2): 92,  # PTX 9.2
}


@lru_cache(maxsize=1)
def get_max_ptx_version() -> int | None:
    """Return the highest PTX ISA version (as ``+ptxNN`` integer) the installed
    CUDA toolkit can assemble, or ``None`` if the toolkit version is not in the
    lookup table.

    When ``None`` is returned the caller should fall back to the NVPTX backend
    default (minimum PTX for the target SM).
    """
    ctk = get_cuda_runtime_version()
    return _CTK_TO_MAX_PTX.get(ctk)


@overload
def get_gpu_compute_capability(as_type: type = str) -> str: ...


@overload
def get_gpu_compute_capability(as_type: type = tuple) -> tuple[int, int]: ...


_cached_cc: tuple[int, int] | None = None


def get_gpu_compute_capability(as_type: type = str) -> str | tuple[int, int]:
    """
    Query the compute capability of the current CUDA device.

    Uses the numba-cuda driver layer so the primary context is shared with
    ``to_device()`` and other device operations.
    """
    global _cached_cc
    assert as_type in (str, tuple), "as_type must be str or tuple"

    if _cached_cc is not None:
        if as_type is tuple:
            return _cached_cc
        return f"sm_{_cached_cc[0]}{_cached_cc[1]}"

    from numba_cuda_mlir.numba_cuda.cudadrv.devices import get_context

    ctx = get_context()
    cc = ctx.device.compute_capability
    _cached_cc = cc
    if as_type is tuple:
        return cc
    return format_arch(cc)


def _check_cuda_result(result):
    """Unwrap CUDA driver API result, raising on error."""
    from cuda.bindings import driver

    if result[0].value != 0:
        _, name = driver.cuGetErrorName(result[0])
        raise RuntimeError(f"CUDA error: {name}")
    return result[1] if len(result) == 2 else result[1:] if len(result) > 2 else None


@lru_cache(maxsize=32)
def get_max_active_clusters(cluster_size: int, device_id: int = 0) -> int:
    """
    Query the maximum number of active clusters for a given cluster size.

    This compiles a minimal dummy kernel and uses cuOccupancyMaxActiveClusters
    to determine the hardware limit for concurrent cluster execution.

    Args:
        cluster_size: Number of CTAs per cluster (1-32)
        device_id: CUDA device ID (default: 0)

    Returns:
        Maximum number of clusters that can be active concurrently
    """
    if cluster_size <= 0 or cluster_size > 32:
        raise ValueError(f"Cluster size must be between 1 and 32, got {cluster_size}")

    from cuda.bindings import driver

    device = _check_cuda_result(driver.cuDeviceGet(device_id))
    max_smem = _check_cuda_result(
        driver.cuDeviceGetAttribute(
            driver.CUdevice_attribute.CU_DEVICE_ATTRIBUTE_MAX_SHARED_MEMORY_PER_BLOCK_OPTIN,
            device,
        )
    )

    kernel = _get_dummy_kernel_function()
    _check_cuda_result(
        driver.cuFuncSetAttribute(
            kernel,
            driver.CUfunction_attribute.CU_FUNC_ATTRIBUTE_MAX_DYNAMIC_SHARED_SIZE_BYTES,
            max_smem,
        )
    )
    max_dyn_smem = _check_cuda_result(driver.cuOccupancyAvailableDynamicSMemPerBlock(kernel, 1, 1))
    max_active_blocks = _check_cuda_result(
        driver.cuOccupancyMaxActiveBlocksPerMultiprocessor(kernel, 1, max_dyn_smem)
    )
    _check_cuda_result(
        driver.cuFuncSetAttribute(
            kernel,
            driver.CUfunction_attribute.CU_FUNC_ATTRIBUTE_NON_PORTABLE_CLUSTER_SIZE_ALLOWED,
            1,
        )
    )

    cluster_dims_attr = driver.CUlaunchAttribute()
    cluster_dims_attr.id = driver.CUlaunchAttributeID.CU_LAUNCH_ATTRIBUTE_CLUSTER_DIMENSION
    (
        cluster_dims_attr.value.clusterDim.x,
        cluster_dims_attr.value.clusterDim.y,
        cluster_dims_attr.value.clusterDim.z,
    ) = (cluster_size, 1, 1)

    launch_config = driver.CUlaunchConfig()
    launch_config.blockDimX, launch_config.blockDimY, launch_config.blockDimZ = (
        128,
        1,
        1,
    )
    launch_config.gridDimX, launch_config.gridDimY, launch_config.gridDimZ = (
        cluster_size,
        max_active_blocks,
        1,
    )
    launch_config.sharedMemBytes, launch_config.numAttrs, launch_config.attrs = (
        max_dyn_smem,
        1,
        [cluster_dims_attr],
    )

    return _check_cuda_result(driver.cuOccupancyMaxActiveClusters(kernel, launch_config))


@lru_cache(maxsize=1)
def _get_dummy_kernel_function():
    """Compile and cache a minimal dummy kernel for occupancy queries."""
    from cuda.bindings import driver
    from numba_cuda_mlir.compiler import compile_cubin
    from numba_cuda_mlir import types

    def _dummy_kernel() -> types.void:
        pass

    cubin = compile_cubin(_dummy_kernel, ())
    cuda_library = _check_cuda_result(driver.cuLibraryLoadData(cubin, None, None, 0, None, None, 0))
    kernels = _check_cuda_result(driver.cuLibraryEnumerateKernels(1, cuda_library))
    return _check_cuda_result(driver.cuKernelGetFunction(kernels[0]))


@lru_cache(maxsize=1)
def is_using_llvm70() -> bool:
    """Return True if the current environment will use the LLVM70 compilation path."""
    from numba_cuda_mlir.mlir_optimization import _needs_llvm70_path

    cc = get_gpu_compute_capability().replace("sm_", "")
    return _needs_llvm70_path(cc)


@lru_cache(maxsize=1)
def get_llvm70_capi_path() -> str:
    """Resolve path to the MLIRToLLVM70 shared library."""
    import numba_cuda_mlir._mlir._mlir_libs as _mlir_libs

    lib_name = "MLIRToLLVM70.dll" if os.name == "nt" else "libMLIRToLLVM70.so"
    capi_path = Path(_mlir_libs.__path__[0]) / lib_name
    if capi_path.exists():
        return str(capi_path.resolve())
    raise FileNotFoundError(
        f"{lib_name} not found under {_mlir_libs.__path__[0]}. "
        "Rebuild numba_cuda_mlir with MLIR_DIR env var set."
    )


def generate_mangled_name(func_name, argtypes):
    """
    Return a mangled name given a function name and argtypes using Numba internals.
    """
    normalized_argtypes = [argtype_normalization(argtype) for argtype in argtypes]
    return itanium_mangler.mangle(func_name, normalized_argtypes)


def argtype_normalization(argtype):
    """
    Normalize the Numba data type
    """
    if isinstance(argtype, types.BooleanLiteral):
        return types.boolean
    elif isinstance(argtype, types.IntegerLiteral):
        return types.int64
    elif isinstance(argtype, types.UniTuple):
        if isinstance(argtype.dtype, types.IntegerLiteral):
            return types.UniTuple(dtype=argtype_normalization(argtype.dtype), count=argtype.count)
        else:
            return argtype
    elif isinstance(argtype, types.NumberClass):
        return argtype.dtype
    else:
        return argtype


def generate_libdevice_stubs():
    print(
        """# SPDX-FileCopyrightText: Copyright (c) 2025 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: BSD-2-Clause

from numba_cuda_mlir.numba_cuda.types import (
    int16,
    int32,
    int64,
    float32,
    float64,
    UniTuple,
    Tuple,
)
"""
    )

    T = '''
def {name}({params}) -> {return_type}:
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/__nv_{name}.html

    CAPI:

    {capi};
    """
'''
    from numba_cuda_mlir.cuda.libdevicefuncs import libdevice_descriptors

    for descriptor in libdevice_descriptors():
        print(descriptor)



# SPDX-FileCopyrightText: Copyright (c) 2025 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: BSD-2-Clause

"""CPython C-API helpers.

The bulk of this module - the ``PythonAPI`` class (PyObject C-API codegen),
``ObjModeUtils``, the box/unbox/reflect contexts and ``EnvironmentManager`` -
built llvmlite IR for object-mode boxing/unboxing and ran only on the vendored
``Lower`` path, which is dead on the MLIR path (device code cannot call the
CPython C-API). It has been removed. What remains is the live extension surface:
the ``PY_UNICODE_*_KIND`` constants (used by the MLIR unicode lowering), the
boxing/unboxing/reflection registries and their ``box``/``unbox``/``reflect``
decorators (the public extension API), and ``NativeValue`` (returned by
extension unbox implementations).
"""

from numba_cuda_mlir import _helperlib
from numba_cuda_mlir.numba_cuda.utils import PYVERSION
from numba_cuda_mlir.numba_cuda import types, cgutils

PY_UNICODE_1BYTE_KIND = _helperlib.py_unicode_1byte_kind
PY_UNICODE_2BYTE_KIND = _helperlib.py_unicode_2byte_kind
PY_UNICODE_4BYTE_KIND = _helperlib.py_unicode_4byte_kind
if PYVERSION in ((3, 10), (3, 11)):
    PY_UNICODE_WCHAR_KIND = _helperlib.py_unicode_wchar_kind


class _Registry:
    def __init__(self):
        self.functions = {}

    def register(self, typeclass):
        assert issubclass(typeclass, types.Type)

        def decorator(func):
            if typeclass in self.functions:
                raise KeyError("duplicate registration for %s" % (typeclass,))
            self.functions[typeclass] = func
            return func

        return decorator

    def lookup(self, typeclass, default=None):
        assert issubclass(typeclass, types.Type)
        for cls in typeclass.__mro__:
            func = self.functions.get(cls)
            if func is not None:
                return func
        return default


# Registries of boxing / unboxing implementations
_boxers = _Registry()
_unboxers = _Registry()
_reflectors = _Registry()

box = _boxers.register
unbox = _unboxers.register
reflect = _reflectors.register


class NativeValue:
    """
    Encapsulate the result of converting a Python object to a native value,
    recording whether the conversion was successful and how to cleanup.
    """

    def __init__(self, value, is_error=None, cleanup=None):
        self.value = value
        self.is_error = is_error if is_error is not None else cgutils.false_bit
        self.cleanup = cleanup



# SPDX-FileCopyrightText: Copyright (c) 2026 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
import math

from numba_cuda_mlir.numba_cuda.typing.templates import (
    AbstractTemplate,
    AttributeTemplate,
    Registry,
    signature,
)
from numba_cuda_mlir.numba_cuda import types
from numba_cuda_mlir.numba_cuda.types.ext_types import Bfloat16

registry = Registry()

# Mapping of function names to their templates
_math_functions = {}


def _make_unary_math_template(key, return_type_fn=None):
    """
    Create a typing template for unary math functions.

    Args:
        key: The function to register (e.g., math.sin)
        return_type_fn: Optional function to compute return type from arg type.
                       If None, returns the same type as input.
    """

    acceptable = (types.Integer, types.Float, Bfloat16)

    class UnaryMathTemplate(AbstractTemplate):
        def generic(self, args, kws):
            if len(args) == 1 and isinstance(args[0], acceptable):
                return_type = return_type_fn(args[0]) if return_type_fn else args[0]
                return signature(return_type, args[0])

    UnaryMathTemplate.key = key
    func_name = key.__name__
    template = registry.register(UnaryMathTemplate)
    _math_functions[func_name] = template
    # Also register as global so the function object can be looked up directly
    registry.register_global(key, types.Function(template))
    return template


def _float_to_integer_return_type(arg_type):
    if isinstance(arg_type, types.Integer):
        return arg_type
    elif arg_type in (types.float16, types.float32):
        return types.int32
    else:
        return types.int64


# Functions that return integer (matching Python 3 behavior)
for func in [math.ceil, math.floor, math.trunc]:
    _make_unary_math_template(func, _float_to_integer_return_type)


# Functions that return the same type as input
for func in [
    math.sin,
    math.cos,
    math.tan,
    math.sqrt,
    math.exp,
    math.expm1,
    math.log,
    math.log2,
    math.log10,
    math.log1p,
    math.fabs,
    math.tanh,
    math.sinh,
    math.cosh,
    math.asin,
    math.acos,
    math.atan,
    math.asinh,
    math.acosh,
    math.atanh,
    math.degrees,
    math.radians,
    math.gamma,
    math.lgamma,
    math.erf,
    math.erfc,
]:
    _make_unary_math_template(func)

# Gated for Python 3.11+
if hasattr(math, "exp2"):
    _make_unary_math_template(math.exp2)


# Functions that return boolean
for func in [math.isfinite, math.isnan, math.isinf]:
    _make_unary_math_template(func, lambda _: types.boolean)


def _make_binary_math_template(key, return_type_fn=None):
    """
    Create a typing template for binary math functions.

    Args:
        key: The function to register (e.g., math.atan2)
        return_type_fn: Optional function to compute return type from arg types.
                       If None, returns the type of the first argument.
    """

    class BinaryMathTemplate(AbstractTemplate):
        def generic(self, args, kws):
            if (
                len(args) == 2
                and isinstance(args[0], types.Number)
                and isinstance(args[1], types.Number)
            ):
                return_type = return_type_fn(args[0], args[1]) if return_type_fn else args[0]
                return signature(return_type, args[0], args[1])

    BinaryMathTemplate.key = key
    func_name = key.__name__
    template = registry.register(BinaryMathTemplate)
    _math_functions[func_name] = template
    registry.register_global(key, types.Function(template))
    return template


# Binary math functions that return same type as input
for func in [
    math.atan2,
    math.copysign,
    math.hypot,
    math.fmod,
    math.remainder,
    math.pow,
    math.nextafter,
]:
    _make_binary_math_template(func)


# Special case: frexp returns a tuple (mantissa: float, exponent: int32)
@registry.register
class FrexpTemplate(AbstractTemplate):
    key = math.frexp

    def generic(self, args, kws):
        if len(args) == 1 and isinstance(args[0], types.Float):
            # frexp returns (mantissa, exponent) where mantissa is same type as input
            return_type = types.Tuple([args[0], types.int32])
            return signature(return_type, args[0])


_math_functions["frexp"] = FrexpTemplate
registry.register_global(math.frexp, types.Function(FrexpTemplate))


# Special case: ldexp takes (float, int) and returns float
@registry.register
class LdexpTemplate(AbstractTemplate):
    key = math.ldexp

    def generic(self, args, kws):
        if len(args) == 2 and isinstance(args[0], types.Float):
            # ldexp takes (mantissa, exponent) and returns same type as mantissa
            if isinstance(args[1], types.Integer):
                return signature(args[0], args[0], args[1])


_math_functions["ldexp"] = LdexpTemplate
registry.register_global(math.ldexp, types.Function(LdexpTemplate))


# Special case: modf returns a tuple (fractional_part, integer_part)
@registry.register
class ModfTemplate(AbstractTemplate):
    key = math.modf

    def generic(self, args, kws):
        if len(args) == 1 and isinstance(args[0], types.Float):
            # modf returns (fractional, integer) both same type as input
            return_type = types.Tuple([args[0], args[0]])
            return signature(return_type, args[0])


_math_functions["modf"] = ModfTemplate
registry.register_global(math.modf, types.Function(ModfTemplate))


@registry.register_attr
class MathModuleAttributeTemplate(AttributeTemplate):
    """
    Resolve attributes of the math module (e.g., math.sin, math.cos).
    """

    key = types.Module(math)

    def resolve(self, mod, attrname):
        """Resolve math module attributes to their function templates."""
        if attrname in _math_functions:
            # Must use the typing context's merged global (augmented with extending
            # registries, e.g. cuDF Masked unary ops). Returning types.Function from
            # _math_functions alone drops extra templates and breaks sin(Masked(float64))
            # when the attribute is resolved via this template before globals merge.
            fn = getattr(math, attrname, None)
            if fn is not None:
                gty = self.context._lookup_global(fn)
                if gty is not None:
                    return gty
            return types.Function(_math_functions[attrname])
        return None


@registry.register_attr
class ComplexAttributeTemplate(AttributeTemplate):
    """
    Resolve attributes of complex numbers (e.g., complex64.real, complex128.imag).
    """

    key = types.Complex

    def resolve_real(self, val):
        """Resolve .real attribute to return the underlying float type."""
        if isinstance(val, types.Complex):
            # complex64 -> float32, complex128 -> float64
            return val.underlying_float
        return None

    def resolve_imag(self, val):
        """Resolve .imag attribute to return the underlying float type."""
        if isinstance(val, types.Complex):
            # complex64 -> float32, complex128 -> float64
            return val.underlying_float
        return None



#  Part of the LLVM Project, under the Apache License v2.0 with LLVM Exceptions.
#  See https://llvm.org/LICENSE.txt for license information.
#  SPDX-License-Identifier: Apache-2.0 WITH LLVM-exception

from ._pdl_ops_gen import *
from ._pdl_ops_gen import _Dialect
from .._mlir_libs._mlirDialectsPDL import *
from .._mlir_libs._mlirDialectsPDL import OperationType
from ..extras.meta import region_op

try:
    from ..ir import *
    from ..dialects import pdl
except ImportError as e:
    raise RuntimeError("Error loading imports from extension module") from e

from typing import Union, Optional, Sequence, Mapping, NewType
from ._ods_common import (
    get_op_result_or_value as _get_value,
    get_op_results_or_values as _get_values,
    _cext as _ods_cext,
)


@_ods_cext.register_operation(_Dialect, replace=True)
class AttributeOp(AttributeOp):
    """Specialization for PDL attribute op class."""

    def __init__(
        self,
        valueType: Optional[Union[OpView, Operation, Value]] = None,
        value: Optional[Attribute] = None,
        *,
        loc=None,
        ip=None,
    ):
        valueType = valueType if valueType is None else _get_value(valueType)
        result = pdl.AttributeType.get()
        super().__init__(result, valueType=valueType, value=value, loc=loc, ip=ip)


@_ods_cext.register_operation(_Dialect, replace=True)
class OperandOp(OperandOp):
    """Specialization for PDL operand op class."""

    def __init__(
        self,
        type: Optional[Union[OpView, Operation, Value]] = None,
        *,
        loc=None,
        ip=None,
    ):
        type = type if type is None else _get_value(type)
        result = pdl.ValueType.get()
        super().__init__(result, valueType=type, loc=loc, ip=ip)


@_ods_cext.register_operation(_Dialect, replace=True)
class OperandsOp(OperandsOp):
    """Specialization for PDL operands op class."""

    def __init__(
        self,
        types: Optional[Union[OpView, Operation, Value]] = None,
        *,
        loc=None,
        ip=None,
    ):
        types = types if types is None else _get_value(types)
        result = pdl.RangeType.get(pdl.ValueType.get())
        super().__init__(result, valueType=types, loc=loc, ip=ip)


@_ods_cext.register_operation(_Dialect, replace=True)
class OperationOp(OperationOp):
    """Specialization for PDL operand op class."""

    def __init__(
        self,
        name: Optional[Union[str, StringAttr]] = None,
        args: Optional[Sequence[Union[OpView, Operation, Value]]] = None,
        attributes: Optional[Mapping[str, Union[OpView, Operation, Value]]] = None,
        types: Optional[Sequence[Union[OpView, Operation, Value]]] = None,
        *,
        loc=None,
        ip=None,
    ):
        if types is None:
            types = []
        if attributes is None:
            attributes = {}
        if args is None:
            args = []
        args = _get_values(args)
        attrNames = []
        attrValues = []
        for attrName, attrValue in attributes.items():
            attrNames.append(StringAttr.get(attrName))
            attrValues.append(_get_value(attrValue))
        attrNames = ArrayAttr.get(attrNames)
        types = _get_values(types)
        result = pdl.OperationType.get()
        super().__init__(
            result, args, attrValues, attrNames, types, opName=name, loc=loc, ip=ip
        )


@_ods_cext.register_operation(_Dialect, replace=True)
class PatternOp(PatternOp):
    """Specialization for PDL pattern op class."""

    def __init__(
        self,
        benefit: Union[IntegerAttr, int],
        name: Optional[Union[StringAttr, str]] = None,
        *,
        loc=None,
        ip=None,
    ):
        """Creates an PDL `pattern` operation."""
        super().__init__(benefit, sym_name=name, loc=loc, ip=ip)
        self.regions[0].blocks.append()

    @property
    def body(self):
        """Return the body (block) of the pattern."""
        return self.regions[0].blocks[0]


pattern = region_op(PatternOp.__base__)


@_ods_cext.register_operation(_Dialect, replace=True)
class ReplaceOp(ReplaceOp):
    """Specialization for PDL replace op class."""

    def __init__(
        self,
        op: Union[OpView, Operation, Value],
        *,
        with_op: Optional[Union[OpView, Operation, Value]] = None,
        with_values: Optional[Sequence[Union[OpView, Operation, Value]]] = None,
        loc=None,
        ip=None,
    ):
        if with_values is None:
            with_values = []
        op = _get_value(op)
        with_op = with_op if with_op is None else _get_value(with_op)
        with_values = _get_values(with_values)
        super().__init__(op, with_values, replOperation=with_op, loc=loc, ip=ip)


@_ods_cext.register_operation(_Dialect, replace=True)
class ResultOp(ResultOp):
    """Specialization for PDL result op class."""

    def __init__(
        self,
        parent: Union[OpView, Operation, Value],
        index: Union[IntegerAttr, int],
        *,
        loc=None,
        ip=None,
    ):
        parent = _get_value(parent)
        result = pdl.ValueType.get()
        super().__init__(result, parent, index, loc=loc, ip=ip)


@_ods_cext.register_operation(_Dialect, replace=True)
class RewriteOp(RewriteOp):
    """Specialization for PDL rewrite op class."""

    def __init__(
        self,
        root: Optional[Union[OpView, Operation, Value]] = None,
        name: Optional[Union[StringAttr, str]] = None,
        args: Optional[Sequence[Union[OpView, Operation, Value]]] = None,
        *,
        loc=None,
        ip=None,
    ):
        if args is None:
            args = []
        root = root if root is None else _get_value(root)
        args = _get_values(args)
        super().__init__(args, root=root, name=name, loc=loc, ip=ip)

    def add_body(self):
        """Add body (block) to the rewrite."""
        self.regions[0].blocks.append()
        return self.body

    @property
    def body(self):
        """Return the body (block) of the rewrite."""
        return self.regions[0].blocks[0]


rewrite = region_op(RewriteOp)


@_ods_cext.register_operation(_Dialect, replace=True)
class TypeOp(TypeOp):
    """Specialization for PDL type op class."""

    def __init__(
        self, constantType: Optional[Union[TypeAttr, Type]] = None, *, loc=None, ip=None
    ):
        result = pdl.TypeType.get()
        super().__init__(result, constantType=constantType, loc=loc, ip=ip)


@_ods_cext.register_operation(_Dialect, replace=True)
class TypesOp(TypesOp):
    """Specialization for PDL types op class."""

    def __init__(
        self,
        constantTypes: Optional[Sequence[Union[TypeAttr, Type]]] = None,
        *,
        loc=None,
        ip=None,
    ):
        if constantTypes is None:
            constantTypes = []
        result = pdl.RangeType.get(pdl.TypeType.get())
        super().__init__(result, constantTypes=constantTypes, loc=loc, ip=ip)


OperationTypeT = NewType("OperationType", OperationType)


def op_t() -> OperationTypeT:
    return OperationTypeT(OperationType.get())



# SPDX-FileCopyrightText: Copyright (c) 2026 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
"""
MLIR NRT context: provides incref/decref emission for the MLIR lowering pass.

Mirrors numba-cuda's NRTContext but emits MLIR LLVM dialect ops instead of
llvmlite IR.  The key entry points are ``incref()`` and ``decref()`` which
walk a type's data model to find all contained NRT MemInfos and emit the
appropriate NRT_incref / NRT_decref calls.

NRT refcounting contract for expression lowerings
--------------------------------------------------
Every Numba IR variable is independently del'd by liveness analysis.  ``del``
calls ``decref``, which walks into contained NRT members via ``traverse_mlir``.
To keep refcounts balanced, every expression lowering (registered builder,
overload impl, getattr impl, etc.) that **stores its result** must follow
this rule:

  * If the expression **creates a fresh NRT object** (e.g. allocates a new
    managed_udf_string via a shim call), the birth gives rc=1.  That single
    reference belongs to the target variable.  **No incref needed.**

  * If the expression **copies or extracts an existing NRT reference** from
    one Numba IR variable into another (struct wrapping, struct field
    extraction, pass-through alias), it must call ``builder.incref(typ, val)``
    on the result before ``store_var``.  Without this, two variables alias the
    same inner NRT pointer, both get del'd, and the second decref is a
    use-after-free.

This matches the Numba CPU convention: ``lower_expr`` in
``numba.core.lowering`` increfs results of ``build_tuple``, ``cast``,
``pair_first``/``pair_second``, and variable-to-variable assignments.
Function call results are NOT incref'd by the framework — the impl/builder
is responsible.

Examples of expressions that NEED incref (copiers/extractors):
  - Struct constructor wrapping an existing NRT member (e.g. NRTWrapperType(inner))
  - Pass-through alias (e.g. pack_return just forwarding a value)
  - Field extraction from a struct (e.g. getattr .inner on NRTWrapperType)
  - Tuple construction from existing NRT elements (handled by numba_cuda_mlir core)
  - Cast of an NRT-managed value (handled by numba_cuda_mlir core)

Examples that do NOT need incref (producers):
  - Shim calls that allocate new NRT objects (concat, upper, lower, etc.)
  - Any operation whose result contains only freshly-born NRT pointers
"""

from numba_cuda_mlir._mlir import ir
from numba_cuda_mlir._mlir.dialects import func, llvm
from numba_cuda_mlir.lowering_utilities import get_or_insert_function
from numba_cuda_mlir.types import BaseTuple


def _get_or_declare_nrt_func(module, name):
    """Declare ``void @name(ptr)`` in *module* if not already present."""

    func_type = ir.FunctionType.get([llvm.PointerType.get()], [])
    return get_or_insert_function(name, func_type, module)


def _emit_nrt_call(module, funcname, meminfo_val):
    """Emit a call to NRT_incref or NRT_decref on a single meminfo pointer."""
    callee = _get_or_declare_nrt_func(module, funcname)
    func.call(result=[], callee=callee.name.value, operands_=[meminfo_val])


class MLIRNRTContext:
    """NRT helpers for the MLIR lowering pass.

    Parameters
    ----------
    data_model_manager : DataModelManager
        The numba_cuda_mlir data model manager used to look up models by Numba type.
    """

    def __init__(self, data_model_manager):
        self._dmm = data_model_manager

    # ------------------------------------------------------------------
    # Public API used by MLIRLower
    # ------------------------------------------------------------------

    def incref(self, module, typ, value):
        """Emit NRT_incref for every MemInfo reachable from *value*."""
        for mi in self._get_meminfos(typ, value):
            _emit_nrt_call(module, "NRT_incref", mi)

    def decref(self, module, typ, value):
        """Emit NRT_decref for every MemInfo reachable from *value*."""
        for mi in self._get_meminfos(typ, value):
            _emit_nrt_call(module, "NRT_decref", mi)

    def type_has_nrt_meminfo(self, typ):
        """Return True if *typ* (or any contained type) needs NRT tracking."""
        try:
            model = self._dmm.lookup(typ)
        except KeyError:
            return False
        if model.has_nrt_meminfo():
            return True
        if isinstance(typ, BaseTuple):
            return any(self.type_has_nrt_meminfo(t) for t in typ)
        if hasattr(model, "traverse_mlir"):
            for member_typ, _getter in model.traverse_mlir():
                if self.type_has_nrt_meminfo(member_typ):
                    return True
        if model.contains_nrt_meminfo():
            return True
        return False

    # ------------------------------------------------------------------
    # Internal helpers
    # ------------------------------------------------------------------

    def _get_meminfos(self, typ, value):
        """Yield all MemInfo pointer values reachable from *value*.

        Walks the data model tree exactly like numba-cuda's
        ``NRTContext.get_meminfos``.
        """
        try:
            model = self._dmm.lookup(typ)
        except KeyError:
            return

        if model.has_nrt_meminfo():
            mi = model.get_nrt_meminfo(value)
            if mi is not None:
                yield mi

        if hasattr(model, "traverse_mlir"):
            for member_typ, getter in model.traverse_mlir():
                member_val = getter(value)
                yield from self._get_meminfos(member_typ, member_val)



# SPDX-FileCopyrightText: Copyright (c) 2026 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
from importlib.util import find_spec
from numba_cuda_mlir._mlir import ir
from textwrap import indent
from pathlib import Path
from io import StringIO
import re
from textwrap import dedent
import inspect
import os
from numba_cuda_mlir._mlir.ir import Module, Operation
from typing import Iterable, Union
import sys

# For legacy Numba-CUDA filecheck test cases
from filecheck.matcher import Matcher
from filecheck.options import Options
from filecheck.parser import Parser, pattern_for_opts
from filecheck.finput import FInput
from numba_cuda_mlir.descriptor import MLIRDispatcher
import cmath
import datetime
import enum
import math
import numpy as np
import pytest
import subprocess
import tempfile
import unittest

np_version = tuple(map(int, np.__version__.split(".")[:2]))
IS_NUMPY_2 = np_version >= (2, 0)


def captured_output(capfd) -> str:
    captured = capfd.readouterr()
    return captured.out + captured.err


def run_in_subprocess(code: str, timeout: int = 60) -> tuple[str, str]:
    """Run Python code in a subprocess and return stdout/stderr."""
    tmp = tempfile.NamedTemporaryFile(mode="w", suffix=".py", encoding="utf-8", delete=False)
    try:
        tmp.write(dedent(code))
        tmp.close()
        proc = subprocess.run(
            [sys.executable, "-u", tmp.name],
            timeout=timeout,
            capture_output=True,
            check=True,
            text=True,
        )
    except subprocess.CalledProcessError as e:
        raise AssertionError(
            f"process failed with code {e.returncode:d}:\n"
            f"stdout:\n{e.stdout}\n\nstderr:\n{e.stderr}"
        ) from e
    finally:
        tmp.close()
        os.unlink(tmp.name)
    return proc.stdout, proc.stderr


def filecheck_with_comments(module):
    if isinstance(module, Module):
        module = module.operation
    if isinstance(module, Operation):
        assert module.verify()

    op = str(module).strip()
    op = "\n".join(filter(None, op.splitlines()))
    op = dedent(op)

    fun = inspect.currentframe().f_back.f_code
    _, lnum = inspect.findsource(fun)
    fun_with_checks = inspect.getsource(fun)
    check_content = ("\n" * lnum + fun_with_checks).replace("\r\n", "\n")

    opts = Options(match_filename="-", check_prefixes=["CHECK"])
    input_file = FInput(fname="-", content=op)
    parser = Parser(opts, StringIO(check_content), *pattern_for_opts(opts))
    matcher = Matcher(opts, input_file, parser)
    matcher.stderr = StringIO()
    try:
        result = matcher.run()
    except AssertionError as e:
        err = matcher.stderr.getvalue()
        raise ValueError(f"\n{err}") from e
    if result != 0:
        err = matcher.stderr.getvalue()
        raise ValueError(f"\n{err}")


def filecheck(checks: str, actual: str | bytes | ir.Module):
    if isinstance(actual, ir.Module):
        actual = str(actual)
    if isinstance(actual, bytes):
        actual = actual.decode()
    checks = dedent(checks).replace("\r\n", "\n")
    actual = actual.replace("\r\n", "\n")

    opts = Options(match_filename="-", check_prefixes=["CHECK"])
    input_file = FInput(fname="-", content=actual)
    parser = Parser(opts, StringIO(checks), *pattern_for_opts(opts))
    matcher = Matcher(opts, input_file, parser)
    matcher.stderr = StringIO()
    result = matcher.run()
    if result != 0:
        err = matcher.stderr.getvalue()
        raise ValueError(f"\n{err}")


# From legacy Numba-CUDA testsuite


@pytest.mark.usefixtures("initialize_from_pytest_config")
class NumbaCUDATestCase(unittest.TestCase):
    """
    For tests copied from the legacy Numba-CUDA testsuite that use FileCheck.

    Method assertFileCheckAsm will inspect an MLIRDispatcher and assert that
    the compilation artifacts match the FileCheck checks given in the kernel's
    docstring.

    Method assertFileCheckMatches can be used to assert that a given string
    matches FileCheck checks, and is not specific to MLIRDispatcher.
    """

    Signature = Union[tuple[type, ...], None]

    _bool_types = (bool, np.bool_)
    _exact_typesets = [_bool_types, (int,), (str,), (np.integer,), (bytes, np.bytes_)]
    _approx_typesets = [(float,), (complex,), (np.inexact)]
    _sequence_typesets = [(tuple, list)]
    _float_types = (float, np.floating)
    _complex_types = (complex, np.complexfloating)

    FLOAT16_RTOL = np.finfo(np.float16).eps

    def _getIRContents(
        self,
        ir_result: Union[dict[Signature, str], str],
        signature: Union[Signature, None] = None,
    ) -> Iterable[str]:
        if isinstance(ir_result, str):
            assert signature is None, (
                "Cannot use signature because the kernel was only compiled for one signature"
            )
            return [ir_result]

        if signature is None:
            return list(ir_result.values())

        return [ir_result[signature]]

    def assertFileCheckAsm(
        self,
        ir_producer: MLIRDispatcher,
        signature: Union[tuple[type, ...], None] = None,
        check_prefixes: tuple[str] = ("ASM",),
        **extra_filecheck_options,
    ) -> None:
        """
        Assert that the assembly output of the given MLIRDispatcher matches
        the FileCheck checks given in the kernel's docstring.
        """
        ir_contents = self._getIRContents(ir_producer.inspect_asm(), signature)
        assert ir_contents, "No assembly output found for the given signature."
        assert ir_producer.__doc__ is not None, (
            "Kernel docstring is required. To pass checks explicitly, use assertFileCheckMatches."
        )
        check_patterns = ir_producer.__doc__
        for ir_content in ir_contents:
            self.assertFileCheckMatches(
                ir_content,
                check_patterns=check_patterns,
                check_prefixes=check_prefixes,
                **extra_filecheck_options,
            )

    def assertFileCheckMatches(
        self,
        ir_content: str,
        check_patterns: str,
        check_prefixes: tuple[str] = ("CHECK",),
        **extra_filecheck_options,
    ) -> None:
        """
        Assert that the given string matches the passed FileCheck checks.

        Args:
            ir_content: The string to check against.
            check_patterns: The FileCheck checks to use.
            check_prefixes: The prefixes to use for the FileCheck checks.
            extra_filecheck_options: Extra options to pass to FileCheck.
        """
        opts = Options(
            match_filename="-",
            check_prefixes=list(check_prefixes),
            **extra_filecheck_options,
        )
        input_file = FInput(fname="-", content=ir_content)
        parser = Parser(opts, StringIO(check_patterns), *pattern_for_opts(opts))
        matcher = Matcher(opts, input_file, parser)
        matcher.stderr = StringIO()
        result = matcher.run()
        if result != 0:
            if self._dump_failed_filechecks:
                dump_directory = Path(datetime.now().strftime("numba-ir-%Y_%m_%d_%H_%M_%S"))
                if not dump_directory.exists():
                    dump_directory.mkdir(parents=True, exist_ok=True)
                base_path = self.id().replace(".", "_")
                ir_dump = dump_directory / Path(base_path).with_suffix(".ll")
                checks_dump = dump_directory / Path(base_path).with_suffix(".checks")
                with (
                    open(ir_dump, "w") as ir_file,
                    open(checks_dump, "w") as checks_file,
                ):
                    _ = ir_file.write(ir_content + "\n")
                    _ = checks_file.write(check_patterns)
                    dump_instructions = f"Reproduce with:\n\nfilecheck --check-prefixes={','.join(check_prefixes)} {checks_dump} --input-file {ir_dump}"
            else:
                dump_instructions = "Rerun with --dump-failed-filechecks to generate a reproducer."

            self.fail(f"FileCheck failed:\n{matcher.stderr.getvalue()}\n\n" + dump_instructions)

    def assertPreciseEqual(
        self,
        first,
        second,
        prec="exact",
        ulps=1,
        msg=None,
        ignore_sign_on_zero=False,
        abs_tol=None,
    ):
        """
        Versatile equality testing function with more built-in checks than
        standard assertEqual().

        For arrays, test that layout, dtype, shape are identical, and
        recursively call assertPreciseEqual() on the contents.

        For other sequences, recursively call assertPreciseEqual() on
        the contents.

        For scalars, test that two scalars or have similar types and are
        equal up to a computed precision.
        If the scalars are instances of exact types or if *prec* is
        'exact', they are compared exactly.
        If the scalars are instances of inexact types (float, complex)
        and *prec* is not 'exact', then the number of significant bits
        is computed according to the value of *prec*: 53 bits if *prec*
        is 'double', 24 bits if *prec* is single.  This number of bits
        can be lowered by raising the *ulps* value.
        ignore_sign_on_zero can be set to True if zeros are to be considered
        equal regardless of their sign bit.
        abs_tol if this is set to a float value its value is used in the
        following. If, however, this is set to the string "eps" then machine
        precision of the type(first) is used in the following instead. This
        kwarg is used to check if the absolute difference in value between first
        and second is less than the value set, if so the numbers being compared
        are considered equal. (This is to handle small numbers typically of
        magnitude less than machine precision).

        Any value of *prec* other than 'exact', 'single' or 'double'
        will raise an error.
        """
        try:
            self._assertPreciseEqual(first, second, prec, ulps, msg, ignore_sign_on_zero, abs_tol)
        except AssertionError as exc:
            failure_msg = str(exc)
            # Fall off of the 'except' scope to avoid Python 3 exception
            # chaining.
        else:
            return
        # Decorate the failure message with more information
        self.fail("when comparing %s and %s: %s" % (first, second, failure_msg))

    def _assertPreciseEqual(
        self,
        first,
        second,
        prec="exact",
        ulps=1,
        msg=None,
        ignore_sign_on_zero=False,
        abs_tol=None,
    ):
        """Recursive workhorse for assertPreciseEqual()."""

        def _assertNumberEqual(first, second, delta=None):
            if delta is None or first == second == 0.0 or math.isinf(first) or math.isinf(second):
                self.assertEqual(first, second, msg=msg)
                # For signed zeros
                if not ignore_sign_on_zero:
                    try:
                        if math.copysign(1, first) != math.copysign(1, second):
                            self.fail(self._formatMessage(msg, "%s != %s" % (first, second)))
                    except TypeError:
                        pass
            else:
                self.assertAlmostEqual(first, second, delta=delta, msg=msg)

        first_family = self._detect_family(first)
        second_family = self._detect_family(second)

        assertion_message = "Type Family mismatch. (%s != %s)" % (
            first_family,
            second_family,
        )
        if msg:
            assertion_message += ": %s" % (msg,)
        self.assertEqual(first_family, second_family, msg=assertion_message)

        # We now know they are in the same comparison family
        compare_family = first_family

        # For recognized sequences, recurse
        if compare_family == "ndarray":
            dtype = self._fix_dtype(first.dtype)
            self.assertEqual(dtype, self._fix_dtype(second.dtype))
            self.assertEqual(first.ndim, second.ndim, "different number of dimensions")
            self.assertEqual(first.shape, second.shape, "different shapes")
            self.assertEqual(first.flags.writeable, second.flags.writeable, "different mutability")
            # itemsize is already checked by the dtype test above
            self.assertEqual(
                self._fix_strides(first), self._fix_strides(second), "different strides"
            )
            if first.dtype != dtype:
                first = first.astype(dtype)
            if second.dtype != dtype:
                second = second.astype(dtype)
            for a, b in zip(first.flat, second.flat):
                self._assertPreciseEqual(a, b, prec, ulps, msg, ignore_sign_on_zero, abs_tol)
            return

        elif compare_family == "sequence":
            self.assertEqual(len(first), len(second), msg=msg)
            for a, b in zip(first, second):
                self._assertPreciseEqual(a, b, prec, ulps, msg, ignore_sign_on_zero, abs_tol)
            return

        elif compare_family == "exact":
            exact_comparison = True

        elif compare_family in ["complex", "approximate"]:
            exact_comparison = False

        elif compare_family == "enum":
            self.assertIs(first.__class__, second.__class__)
            self._assertPreciseEqual(
                first.value, second.value, prec, ulps, msg, ignore_sign_on_zero, abs_tol
            )
            return

        elif compare_family == "unknown":
            # Assume these are non-numeric types: we will fall back
            # on regular unittest comparison.
            self.assertIs(first.__class__, second.__class__)
            exact_comparison = True

        else:
            assert 0, "unexpected family"

        # If a Numpy scalar, check the dtype is exactly the same too
        # (required for datetime64 and timedelta64).
        if hasattr(first, "dtype") and hasattr(second, "dtype"):
            self.assertEqual(first.dtype, second.dtype)

        # Mixing bools and non-bools should always fail
        if isinstance(first, self._bool_types) != isinstance(second, self._bool_types):
            assertion_message = "Mismatching return types (%s vs. %s)" % (
                first.__class__,
                second.__class__,
            )
            if msg:
                assertion_message += ": %s" % (msg,)
            self.fail(assertion_message)

        try:
            if cmath.isnan(first) and cmath.isnan(second):
                # The NaNs will compare unequal, skip regular comparison
                return
        except TypeError:
            # Not floats.
            pass

        # if absolute comparison is set, use it
        if abs_tol is not None:
            if abs_tol == "eps":
                rtol = np.finfo(type(first)).eps
            elif isinstance(abs_tol, float):
                rtol = abs_tol
            else:
                raise ValueError('abs_tol is not "eps" or a float, found %s' % abs_tol)
            if abs(first - second) < rtol:
                return

        exact_comparison = exact_comparison or prec == "exact"

        if not exact_comparison and prec != "exact":
            if prec == "single":
                bits = 24
            elif prec == "double":
                bits = 53
            else:
                raise ValueError("unsupported precision %r" % (prec,))
            k = 2 ** (ulps - bits - 1)
            delta = k * (abs(first) + abs(second))
        else:
            delta = None
        if isinstance(first, self._complex_types):
            _assertNumberEqual(first.real, second.real, delta)
            _assertNumberEqual(first.imag, second.imag, delta)
        elif isinstance(first, (np.timedelta64, np.datetime64)):
            # Since Np 1.16 NaT == NaT is False, so special comparison needed
            if np.isnat(first):
                self.assertEqual(np.isnat(first), np.isnat(second))
            else:
                _assertNumberEqual(first, second, delta)
        else:
            _assertNumberEqual(first, second, delta)

    def _detect_family(self, numeric_object):
        """
        This function returns a string description of the type family
        that the object in question belongs to.  Possible return values
        are: "exact", "complex", "approximate", "sequence", and "unknown"
        """
        if isinstance(numeric_object, np.ndarray):
            return "ndarray"

        if isinstance(numeric_object, enum.Enum):
            return "enum"

        for tp in self._sequence_typesets:
            if isinstance(numeric_object, tp):
                return "sequence"

        for tp in self._exact_typesets:
            if isinstance(numeric_object, tp):
                return "exact"

        for tp in self._complex_types:
            if isinstance(numeric_object, tp):
                return "complex"

        for tp in self._approx_typesets:
            if isinstance(numeric_object, tp):
                return "approximate"

        return "unknown"

    def _fix_dtype(self, dtype):
        """
        Fix the given *dtype* for comparison.
        """
        # Under 64-bit Windows, Numpy may return either int32 or int64
        # arrays depending on the function.
        if sys.platform == "win32" and sys.maxsize > 2**32 and dtype == np.dtype("int32"):
            return np.dtype("int64")
        else:
            return dtype

    def _fix_strides(self, arr):
        """
        Return the strides of the given array, fixed for comparison.
        Strides for 0- or 1-sized dimensions are ignored.
        """
        if arr.size == 0:
            return [0] * arr.ndim
        else:
            return [
                stride / arr.itemsize
                for (stride, shape) in zip(arr.strides, arr.shape)
                if shape > 1
            ]



# SPDX-FileCopyrightText: Copyright (c) 2026 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
from numba_cuda_mlir._mlir import ir
import contextvars
from typing import Any

_context: ir.Context | None = None
_compilation_options: contextvars.ContextVar[dict[str, Any] | None] = contextvars.ContextVar(
    "compilation_options", default=None
)


def get_context() -> ir.Context:
    global _context
    if _context is None:
        _context = ir.Context()
    return _context


def get_compilation_options() -> dict[str, Any]:
    opts = _compilation_options.get()
    if opts is None:
        opts = {}
        _compilation_options.set(opts)
    return opts


def set_compilation_options(options: dict[str, Any]) -> contextvars.Token:
    return _compilation_options.set(options)



#  Part of the LLVM Project, under the Apache License v2.0 with LLVM Exceptions.
#  See https://llvm.org/LICENSE.txt for license information.
#  SPDX-License-Identifier: Apache-2.0 WITH LLVM-exception
"""Represents configured ops as emitted for code generation.

Classes in this module generally are directly serializable to YAML for use
by the code generator.

TODO: These should just be dumb containers or serialization code but they
currently encode too many details of how the language is interpreted. Move this
to helpers on the comprehension objects themselves.
"""

from typing import Dict, Optional

from ..... import ir as _ir
from .comprehension import *
from .yaml_helper import *

__all__ = ["LinalgStructuredOpConfig", "LinalgOpConfig", "OperandDefConfig"]


def _serialize_affine_map(affine_map: _ir.AffineMap) -> str:
    with affine_map.context:
        # Affine map printing/parsing is via an AffineMap attr.
        attr = _ir.AffineMapAttr.get(affine_map)
        return str(attr)


class TensorUseConfig:
    """Wrapper around a TensorUse with additional context-bound state."""

    def __init__(self, tensor_use: TensorUse, indexing_map: _ir.AffineMap):
        self.tensor_use = tensor_use
        self.indexing_map = indexing_map

    def __repr__(self):
        return f"Use({self.tensor_use}, indexing_map={self.indexing_map})"


class OperandDefConfig(YAMLObject):
    """Wrapper containing an operand definition with additional state."""

    yaml_tag = "!LinalgOperandDefConfig"

    def __init__(
        self,
        operand_def: OperandDef,
        shape_map: Optional[_ir.AffineMap] = None,
        index_attr_map: Optional[_ir.AffineMap] = None,
    ):
        self.operand_def = operand_def
        self.shape_map = shape_map  # type: Optional[_ir.AffineMap]
        self.index_attr_map = index_attr_map  # type: Optional[_ir.AffineMap]
        self.indexing_map = None  # type: Optional[_ir.AffineMap]

    @property
    def name(self) -> str:
        return self.operand_def.name

    @property
    def kind(self) -> OperandKind:
        return self.operand_def.kind

    @property
    def type_var(self) -> TypeVar:
        return self.operand_def.type_var

    def to_yaml_custom_dict(self):
        self_dict = dict(name=self.name, kind=self.operand_def.kind.name.lower())
        if self.type_var:
            self_dict["type_var"] = self.type_var.name
        if self.shape_map:
            self_dict["shape_map"] = _serialize_affine_map(self.shape_map)
        if self.index_attr_map:
            self_dict["index_attr_map"] = _serialize_affine_map(self.index_attr_map)
        if self.operand_def.default_indices:
            self_dict["default_indices"] = self.operand_def.default_indices
        if self.operand_def.default_fn:
            self_dict["default_fn"] = self.operand_def.default_fn
        return self_dict

    def __repr__(self):
        return (
            f"OperandDefConfig({self.operand_def}, "
            f"shape_map={self.shape_map}, "
            f"index_attr_map={self.index_attr_map}, "
            f"indexing_map={self.indexing_map})"
        )


class LinalgIndexingMapsConfig(YAMLObject):
    """Abstracts the style of indexing maps that the op exports.

    Presently only static (tied to the op name) indexing maps are supported. In
    the future, it is expected that we will have additional variants:
      - Dynamic based on attributes
      - Dynamic based on operands
    Each is expected to require a different variant of specification.
    """

    yaml_tag = "!LinalgIndexingMapsConfig"

    def __init__(self, static_indexing_maps: Optional[Sequence[_ir.AffineMap]] = None):
        self.static_indexing_maps = static_indexing_maps

    def to_yaml_custom_dict(self):
        if self.static_indexing_maps is not None:
            return dict(
                static_indexing_maps=[
                    _serialize_affine_map(m) for m in self.static_indexing_maps
                ]
            )
        raise ValueError(
            f"LinalgIndexingMapsConfig must have one type of indexing map" f"(got none)"
        )


class LinalgStructuredOpConfig(YAMLObject):
    """Configuration for metadata sufficient to construct a linalg named op."""

    yaml_tag = "!LinalgStructuredOpConfig"

    def __init__(
        self,
        comprehension: Comprehension,
        domain: Sequence[DimDef],
        registered_operands: Sequence[OperandDef],
        context: Optional[_ir.Context] = None,
    ):
        self.context = context if context is not None else _ir.Context()
        self.affine_state = AffineBuildState()
        self.writes = list()  # type: List[Tuple[TensorUse, TensorExpression]]
        self.operands = dict()  # type: Dict[OperandDef, OperandDefConfig]
        self.uses = dict()  # type: Dict[TensorUse, TensorUseConfig]

        # Compute the ordered set of writes and collect the tensor, capture, dims,
        # and index uses.
        collected_tensor_uses = set()
        collected_scalar_uses = set()
        collected_dim_uses = set()
        collected_indices = set()
        for write_use, read_use in zip(comprehension.definitions, comprehension.values):
            self.writes.append((write_use, read_use))

        for write_use, read_use in self.writes:
            collected_tensor_uses.add(write_use)
            read_use.collect_tensor_uses(collected_tensor_uses)
            read_use.collect_scalar_uses(collected_scalar_uses)
            read_use.collect_dim_uses(collected_dim_uses)
            write_use.collect_dim_uses(collected_dim_uses)
            read_use.collect_indices(collected_indices)

        # Set domain to the sorted list of uses if no domain annotation is given.
        if not domain:
            domain = sorted(collected_dim_uses, key=lambda dim: dim.dimname)

        # Verify the domain dimensions match the used dimensions.
        if len(domain) != len(collected_dim_uses) or any(
            dim not in collected_dim_uses for dim in domain
        ):
            raise ValueError(
                f"Expected the annotated domain dimensions {domain} to "
                f"match the set of dimension used by the tensor "
                f"comprehension {collected_dim_uses}"
            )

        # Instantiate the dimensions in the given order.
        with self.context:
            local_state = AffineBuildState(
                global_state=self.affine_state, allow_new_symbols=False
            )
            for dim in domain:
                dim.build(state=local_state)

        # Collect all attribute definitions.
        collected_attr_defs = list()
        for operand in registered_operands:
            if operand.is_attribute():
                collected_attr_defs.append(operand)

        # Collect all tensors with manual indexing annotation.
        collected_index_defs = list()
        for operand in registered_operands:
            if operand.index_dims:
                if any(dim not in collected_dim_uses for dim in operand.index_dims):
                    raise ValueError(
                        f"Expected all index dims {operand.index_dims} of "
                        f"operand {operand.name} to have uses."
                    )
                collected_index_defs.append(operand)

        # Collect the operand definitions of all tensor/scalar uses, attributes, and
        # shape-only tensors.
        all_operand_defs = list()
        for use in collected_tensor_uses:
            all_operand_defs.append(use.operand_def)
        for use in collected_scalar_uses:
            all_operand_defs.append(use.operand_def)
        for definition in collected_attr_defs:
            all_operand_defs.append(definition)
        for definition in collected_index_defs:
            all_operand_defs.append(definition)

        # Add all operands in registration order to ensure the symbols are
        # registered in the order they appear.
        all_operand_defs = sorted(
            all_operand_defs, key=lambda operand_def: operand_def.registered_index
        )
        for operand_def in all_operand_defs:
            self.add_operand(operand_def)

        # Add all shape-only tensor index_dim annotations and all tensor uses.
        for definition in collected_index_defs:
            self.add_indexed_operand(definition)
        for use in collected_tensor_uses:
            self.add_tensor_use(use)

        # Normalize all shape and indexing maps now that full count of dims and
        # symbols are known.
        for cuse in self.uses.values():
            cuse.indexing_map = self._normalize_affine_map(cuse.indexing_map)
        for definition in collected_index_defs:
            self.operands[definition].indexing_map = self._normalize_affine_map(
                self.operands[definition].indexing_map
            )
        for operand_config in self.operands.values():
            if operand_config.shape_map:
                operand_config.shape_map = self._normalize_affine_map(
                    operand_config.shape_map, with_dims=False
                )
            if operand_config.index_attr_map:
                operand_config.index_attr_map = self._normalize_affine_map(
                    operand_config.index_attr_map, with_dims=False
                )

        # Now for each write use, propagate the indexing maps from the use to the
        # tensor, ensuring that there are not conflicts.
        for write_use, _ in self.writes:
            write_tensor_config = self.operands[write_use.operand_def]
            if write_tensor_config.indexing_map:
                raise ValueError(
                    f"Unexpected multi-write to a single tensor: {write_tensor_config}"
                )
            write_tensor_config.indexing_map = self.uses[write_use].indexing_map

        # For each read use, propagate the indexing maps from the use to the
        # tensor, ensuring that there are not conflicts.
        for _, read_expr in self.writes:
            read_uses = set()  # type: Set[TensorUse]
            read_expr.collect_tensor_uses(read_uses)
            for read_use in read_uses:
                read_operand_config = self.operands[read_use.operand_def]
                if (
                    read_operand_config.indexing_map
                    and read_operand_config.indexing_map
                    != self.uses[read_use].indexing_map
                ):
                    raise ValueError(
                        f"Unexpected multi-read of a tensor with different accesses:"
                        f"{read_operand_config} vs {read_use}"
                    )
                read_operand_config.indexing_map = self.uses[read_use].indexing_map

        # Set the indexing map of all scalar uses to the empty map.
        for operand_config in self.operands.values():
            if operand_config.operand_def.kind == OperandKind.SCALAR:
                operand_config.indexing_map = self._get_scalar_map()

        # Check all registered tensor and scalar operands have an indexing map.
        for operand in registered_operands:
            if operand.is_attribute():
                continue
            if not (operand in self.operands and self.operands[operand].indexing_map):
                raise ValueError(
                    f"Failed to compute an indexing map for operand " f"{operand.name}"
                )

        # Collect reduction dims and ensure all the same.
        all_reduction_dims = set(comprehension.all_reduction_dims)
        if len(all_reduction_dims) != 1:
            raise ValueError(
                f"All writes within a generic must have the same reduction "
                f"dims. Got: {all_reduction_dims}"
            )
        self.reduction_dims = next(iter(all_reduction_dims))

        # Check the index dimension exists and resolve.
        for index in collected_indices:
            if index.dim_def.dimname not in self.affine_state.all_dims:
                raise ValueError(
                    f"The dimension {index.dim_def.dimname} is not part of the "
                    f"iteration domain {self.affine_state.all_dims}"
                )
            index.resolve_dimension_name(self.affine_state)

        # Generate the scalar assignments (used to build a body).
        self.assignments = [
            ScalarAssign(write_use.tensor_name, read_expr.to_scalar_expression())
            for write_use, read_expr in self.writes
        ]

    @property
    def ordered_operands(self) -> Sequence[OperandDefConfig]:
        return sorted(
            self.operands.values(),
            key=lambda operand: operand.operand_def.registered_index,
        )

    @property
    def ordered_dims(self) -> Sequence[Tuple[str, int]]:
        """Gets the ordered list of dim bindings (symbolic name, position).

        TODO: The original parser relies on parse ordering to arrive at the
        iterator types, but that ordering is not defined on the Python side, so
        this may be ambiguous.
        """
        return list(self.affine_state.all_dims.items())

    @property
    def indexing_maps(self) -> Sequence[_ir.AffineMap]:
        return [o.indexing_map for o in self.ordered_operands if o.indexing_map]

    @property
    def iterator_types(self) -> Sequence[str]:
        def get_type(symbolic_name, position):
            for reduction_dim_expr in self.reduction_dims:
                if reduction_dim_expr.dimname == symbolic_name:
                    return "reduction"
            return "parallel"

        return [get_type(*dim) for dim in self.ordered_dims]

    def add_operand(self, operand_def: OperandDef):
        if operand_def in self.operands:
            return
        if not (operand_def.is_tensor() or operand_def.kind == OperandKind.INDEX_ATTR):
            self.operands[operand_def] = OperandDefConfig(operand_def)
            return
        with self.context:
            local_state = AffineBuildState(
                global_state=self.affine_state, allow_new_dims=False
            )
            exprs = []
            for expr in operand_def.size_exprs:
                exprs.append(expr.build(state=local_state))
            assert local_state.local_dim_count == 0
            affine_map = _ir.AffineMap.get(
                dim_count=0, symbol_count=local_state.symbol_count, exprs=exprs
            )
            if operand_def.kind == OperandKind.INDEX_ATTR:
                self.operands[operand_def] = OperandDefConfig(
                    operand_def, index_attr_map=affine_map
                )
            else:
                self.operands[operand_def] = OperandDefConfig(
                    operand_def, shape_map=affine_map
                )

    def add_indexed_operand(self, operand_def: OperandDef):
        with self.context:
            local_state = AffineBuildState(
                global_state=self.affine_state, allow_new_symbols=False
            )
            exprs = []
            for expr in operand_def.index_dims:
                exprs.append(expr.build(state=local_state))
            self.operands[operand_def].indexing_map = _ir.AffineMap.get(
                dim_count=local_state.dim_count,
                symbol_count=local_state.symbol_count,
                exprs=exprs,
            )

    def add_tensor_use(self, tensor_use: TensorUse):
        if tensor_use in self.uses:
            return
        with self.context:
            local_state = AffineBuildState(
                global_state=self.affine_state, allow_new_symbols=False
            )
            exprs = []
            for expr in tensor_use.indices:
                exprs.append(expr.build(state=local_state))
            indexing_map = _ir.AffineMap.get(
                dim_count=local_state.dim_count,
                symbol_count=local_state.symbol_count,
                exprs=exprs,
            )

            use_config = TensorUseConfig(tensor_use, indexing_map)
            self.uses[tensor_use] = use_config

    def _get_scalar_map(self) -> _ir.AffineMap:
        """Create an empty affine map used to index a scalar."""
        with self.context:
            return _ir.AffineMap.get(
                dim_count=self.affine_state.dim_count,
                symbol_count=self.affine_state.symbol_count,
                exprs=list(),
            )

    def _normalize_affine_map(
        self, affine_map: _ir.AffineMap, with_dims: bool = True
    ) -> _ir.AffineMap:
        """Normalizes an indexing map to have the max known symbols and dims."""
        with self.context:
            return _ir.AffineMap.get(
                dim_count=self.affine_state.dim_count if with_dims else 0,
                symbol_count=self.affine_state.symbol_count,
                exprs=list(affine_map.results),
            )

    def to_yaml_custom_dict(self):
        self_dict = dict(args=self.ordered_operands)
        # TODO: Refactor the hierarchy internally when supporting more
        # than static (preserving this serialized form).
        self_dict["indexing_maps"] = LinalgIndexingMapsConfig(
            static_indexing_maps=self.indexing_maps
        )
        self_dict["iterator_types"] = self.iterator_types
        self_dict["assignments"] = self.assignments
        return self_dict

    def __repr__(self):
        lines = [f"LinalgGenericOpConfig(reduction_dims={self.reduction_dims},"]
        lines.append("operands=[")
        for def_config in self.ordered_operands:
            lines.append(f"  {repr(def_config)}")
        lines.append("], indexing_maps=[")
        for m in self.indexing_maps:
            lines.append(f"  {repr(m)}")
        lines.append(f"], iterator_types=[")
        for t in self.iterator_types:
            lines.append(f"  {t}")
        lines.append("])")
        return "\n".join(lines)


class LinalgOpConfig(YAMLObject):
    """Container for any supported linalg op type.

    This includes the concrete type by name for ease of parsing by systems
    that ignore tags.
    """

    yaml_tag = "!LinalgOpConfig"

    def __init__(
        self,
        metadata: OpMetadataDef,
        *,
        structured_op: Optional[LinalgStructuredOpConfig] = None,
    ):
        self.metadata = metadata
        self.structured_op = structured_op

    def to_yaml_custom_dict(self):
        self_dict = dict(
            metadata=self.metadata,
        )
        if self.structured_op:
            self_dict["structured_op"] = self.structured_op
        return self_dict

    @staticmethod
    def from_linalg_op_def(
        op_def: LinalgOpDef, context: Optional[_ir.Context] = None
    ) -> Sequence["LinalgOpConfig"]:
        """Expands a LinalgOpDef into corresponding Linalg configured ops."""
        # TODO: Many LinalgOpDef patterns need to expand to multiple generics.
        assert len(op_def.comprehensions) == 1, "Only one comprehension supported"
        return [
            LinalgOpConfig(
                op_def.metadata,
                structured_op=LinalgStructuredOpConfig(
                    op_def.comprehensions[0],
                    op_def.domain,
                    op_def.registered_operands.values(),
                    context,
                ),
            ),
        ]

    def __repr__(self):
        return (
            f"LinalgOpConfig(metadata={self.metadata},\n"
            f"structured_op={self.structured_op})"
        )



# SPDX-FileCopyrightText: Copyright (c) 2025 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: BSD-2-Clause

from abc import abstractmethod, ABCMeta
import itertools
import os
import contextlib
import uuid
import pickle
import hashlib
import errno
import inspect
import tempfile
import sys

import numba_cuda_mlir
from numba_cuda_mlir.numba_cuda.misc.appdirs import AppDirs
from pathlib import Path

from numba_cuda_mlir.numba_cuda.core import config
from numba_cuda_mlir.numba_cuda.serialize import dumps


def _cache_log(msg, *args):
    if config.DEBUG_CACHE:
        msg = msg % args
        print(msg)


class _Cache(metaclass=ABCMeta):
    @property
    @abstractmethod
    def cache_path(self):
        """
        The base filesystem path of this cache (for example its root folder).
        """

    @abstractmethod
    def load_overload(self, sig, target_context):
        """
        Load an overload for the given signature using the target context.
        The saved object must be returned if successful, None if not found
        in the cache.
        """

    @abstractmethod
    def save_overload(self, sig, data):
        """
        Save the overload for the given signature.
        """

    @abstractmethod
    def enable(self):
        """
        Enable the cache.
        """

    @abstractmethod
    def disable(self):
        """
        Disable the cache.
        """

    @abstractmethod
    def flush(self):
        """
        Flush the cache.
        """


class NullCache(_Cache):
    @property
    def cache_path(self):
        return None

    def load_overload(self, sig, target_context):
        pass

    def save_overload(self, sig, data):
        pass

    def enable(self):
        pass

    def disable(self):
        pass

    def flush(self):
        pass


class IndexDataCacheFile:
    """
    Implements the logic for the index file and data file used by a cache.
    """

    def __init__(self, cache_path, filename_base, source_stamp):
        self._cache_path = cache_path
        self._index_name = "%s.nbi" % (filename_base,)
        self._index_path = os.path.join(self._cache_path, self._index_name)
        self._data_name_pattern = "%s.{number:d}.nbc" % (filename_base,)
        self._source_stamp = source_stamp
        self._version = numba_cuda_mlir.__version__

    def flush(self):
        self._save_index({})

    def save(self, key, data):
        """
        Save a new cache entry with *key* and *data*.
        """
        overloads = self._load_index()
        try:
            # If key already exists, we will overwrite the file
            data_name = overloads[key]
        except KeyError:
            # Find an available name for the data file
            existing = set(overloads.values())
            for i in itertools.count(1):
                data_name = self._data_name(i)
                if data_name not in existing:
                    break
            overloads[key] = data_name
            self._save_index(overloads)
        self._save_data(data_name, data)

    def load(self, key):
        """
        Load a cache entry with *key*.
        """
        overloads = self._load_index()
        data_name = overloads.get(key)
        if data_name is None:
            return
        try:
            return self._load_data(data_name)
        except OSError:
            # File could have been removed while the index still refers it.
            return

    def _load_index(self):
        """
        Load the cache index and return it as a dictionary (possibly
        empty if cache is empty or obsolete).
        """
        try:
            with open(self._index_path, "rb") as f:
                version = pickle.load(f)
                data = f.read()
        except FileNotFoundError:
            # Index doesn't exist yet?
            return {}
        if version != self._version:
            # This is another version.  Avoid trying to unpickling the
            # rest of the stream, as that may fail.
            return {}
        stamp, overloads = pickle.loads(data)
        _cache_log("[cache] index loaded from %r", self._index_path)
        if stamp != self._source_stamp:
            # Cache is not fresh.  Stale data files will be eventually
            # overwritten, since they are numbered in incrementing order.
            return {}
        else:
            return overloads

    def _save_index(self, overloads):
        data = self._source_stamp, overloads
        data = self._dump(data)
        with self._open_for_write(self._index_path) as f:
            pickle.dump(self._version, f, protocol=-1)
            f.write(data)
        _cache_log("[cache] index saved to %r", self._index_path)

    def _load_data(self, name):
        path = self._data_path(name)
        with open(path, "rb") as f:
            data = f.read()
        tup = pickle.loads(data)
        _cache_log("[cache] data loaded from %r", path)
        return tup

    def _save_data(self, name, data):
        data = self._dump(data)
        path = self._data_path(name)
        with self._open_for_write(path) as f:
            f.write(data)
        _cache_log("[cache] data saved to %r", path)

    def _data_name(self, number):
        return self._data_name_pattern.format(number=number)

    def _data_path(self, name):
        return os.path.join(self._cache_path, name)

    def _dump(self, obj):
        return dumps(obj)

    @contextlib.contextmanager
    def _open_for_write(self, filepath):
        """
        Open *filepath* for writing in a race condition-free way (hopefully).
        uuid4 is used to try and avoid name collisions on a shared filesystem.
        """
        uid = uuid.uuid4().hex[:16]  # avoid long paths
        tmpname = "%s.tmp.%s" % (filepath, uid)
        try:
            with open(tmpname, "wb") as f:
                yield f
            os.replace(tmpname, filepath)
        except Exception:
            # In case of error, remove dangling tmp file
            try:
                os.unlink(tmpname)
            except OSError:
                pass
            raise


class Cache(_Cache):
    """
    A per-function compilation cache.  The cache saves data in separate
    data files and maintains information in an index file.

    There is one index file per function and Python version
    ("function_name-<lineno>.pyXY.nbi") which contains a mapping of
    signatures and architectures to data files.
    It is prefixed by a versioning key and a timestamp of the Python source
    file containing the function.

    There is one data file ("function_name-<lineno>.pyXY.<number>.nbc")
    per function, function signature, target architecture and Python version.

    Separate index and data files per Python version avoid pickle
    compatibility problems.

    Note:
    This contains the driver logic only.  The core logic is provided
    by a subclass of ``CacheImpl`` specified as *_impl_class* in the subclass.
    """

    # The following class variables must be overridden by subclass.
    _impl_class = None

    def __init__(self, py_func):
        self._name = repr(py_func)
        self._py_func = py_func
        self._impl = self._impl_class(py_func)
        self._cache_path = self._impl.locator.get_cache_path()
        # This may be a bit strict but avoids us maintaining a magic number
        source_stamp = self._impl.locator.get_source_stamp()
        filename_base = self._impl.filename_base
        self._cache_file = IndexDataCacheFile(
            cache_path=self._cache_path,
            filename_base=filename_base,
            source_stamp=source_stamp,
        )
        self.enable()

    def __repr__(self):
        return "<%s py_func=%r>" % (self.__class__.__name__, self._name)

    @property
    def cache_path(self):
        return self._cache_path

    def enable(self):
        self._enabled = True

    def disable(self):
        self._enabled = False

    def flush(self):
        self._cache_file.flush()

    def load_overload(self, sig, target_context):
        """
        Load and recreate the cached object for the given signature,
        using the *target_context*.
        """
        with self._guard_against_spurious_io_errors():
            return self._load_overload(sig, target_context)
        # None returned if the `with` block swallows an exception

    def _load_overload(self, sig, target_context):
        if not self._enabled:
            return
        key = self._index_key(sig, target_context.codegen())
        data = self._cache_file.load(key)
        if data is not None:
            data = self._impl.rebuild(target_context, data)
        return data

    def save_overload(self, sig, data):
        """
        Save the data for the given signature in the cache.
        """
        with self._guard_against_spurious_io_errors():
            self._save_overload(sig, data)

    def _save_overload(self, sig, data):
        if not self._enabled:
            return
        if not self._impl.check_cachable(data):
            return
        self._impl.locator.ensure_cache_path()
        key = self._index_key(sig, data.codegen)
        data = self._impl.reduce(data)
        self._cache_file.save(key, data)

    @contextlib.contextmanager
    def _guard_against_spurious_io_errors(self):
        if os.name == "nt":
            # Guard against permission errors due to accessing the file
            # from several processes (see #2028)
            try:
                yield
            except OSError as e:
                if e.errno != errno.EACCES:
                    raise
        else:
            # No such conditions under non-Windows OSes
            yield

    def _index_key(self, sig, codegen):
        """
        Compute index key for the given signature and codegen.
        It includes a description of the OS, target architecture and hashes of
        the bytecode for the function and, if the function has a __closure__,
        a hash of the cell_contents.
        """
        codebytes = self._py_func.__code__.co_code
        if self._py_func.__closure__ is not None:
            cvars = tuple([x.cell_contents for x in self._py_func.__closure__])
            # Note: cloudpickle serializes a function differently depending
            #       on how the process is launched; e.g. multiprocessing.Process
            cvarbytes = dumps(cvars)
        else:
            cvarbytes = b""

        hasher = lambda x: hashlib.sha256(x).hexdigest()
        return (
            sig,
            codegen.magic_tuple(),
            (
                hasher(codebytes),
                hasher(cvarbytes),
            ),
        )


class _CacheLocator(metaclass=ABCMeta):
    """
    A filesystem locator for caching a given function.
    """

    def ensure_cache_path(self):
        path = self.get_cache_path()
        os.makedirs(path, exist_ok=True)
        # Ensure the directory is writable by trying to write a temporary file
        tempfile.TemporaryFile(dir=path).close()

    @abstractmethod
    def get_cache_path(self):
        """
        Return the directory the function is cached in.
        """

    @abstractmethod
    def get_source_stamp(self):
        """
        Get a timestamp representing the source code's freshness.
        Can return any picklable Python object.
        """

    @abstractmethod
    def get_disambiguator(self):
        """
        Get a string disambiguator for this locator's function.
        It should allow disambiguating different but similarly-named functions.
        """

    @classmethod
    def from_function(cls, py_func, py_file):
        """
        Create a locator instance for the given function located in the
        given file.
        """
        raise NotImplementedError

    @classmethod
    def get_suitable_cache_subpath(cls, py_file):
        """Given the Python file path, compute a suitable path inside the
        cache directory.

        This will reduce a file path that is too long, which can be a problem
        on some operating system (i.e. Windows 7).
        """
        path = os.path.abspath(py_file)
        subpath = os.path.dirname(path)
        parentdir = os.path.split(subpath)[-1]
        # Use SHA1 to reduce path length.
        # Note: windows doesn't like long path.
        hashed = hashlib.sha1(subpath.encode()).hexdigest()
        # Retain parent directory name for easier debugging
        return "_".join([parentdir, hashed])


class _SourceFileBackedLocatorMixin:
    """
    A cache locator mixin for functions which are backed by a well-known
    Python source file.
    """

    def get_source_stamp(self):
        if getattr(sys, "frozen", False):
            st = os.stat(sys.executable)
        else:
            st = os.stat(self._py_file)
        # We use both timestamp and size as some filesystems only have second
        # granularity.
        return st.st_mtime, st.st_size

    def get_disambiguator(self):
        return str(self._lineno)

    @classmethod
    def from_function(cls, py_func, py_file):
        if not os.path.exists(py_file):
            # Perhaps a placeholder (e.g. "<ipython-XXX>")
            return
        self = cls(py_func, py_file)
        try:
            self.ensure_cache_path()
        except OSError:
            # Cannot ensure the cache directory exists or is writable
            return
        return self


class _InTreeCacheLocator(_SourceFileBackedLocatorMixin, _CacheLocator):
    """
    A locator for functions backed by a regular Python module with a
    writable __pycache__ directory.
    """

    def __init__(self, py_func, py_file):
        self._py_file = py_file
        self._lineno = py_func.__code__.co_firstlineno
        self._cache_path = os.path.join(os.path.dirname(self._py_file), "__pycache__")

    def get_cache_path(self):
        return self._cache_path


class _UserProvidedCacheLocator(_SourceFileBackedLocatorMixin, _CacheLocator):
    """
    A locator that always point to the user provided directory in
    `numba.config.CACHE_DIR`
    """

    def __init__(self, py_func, py_file):
        self._py_file = py_file
        self._lineno = py_func.__code__.co_firstlineno
        cache_subpath = self.get_suitable_cache_subpath(py_file)
        self._cache_path = os.path.join(config.CACHE_DIR, cache_subpath)

    def get_cache_path(self):
        return self._cache_path

    @classmethod
    def from_function(cls, py_func, py_file):
        if not config.CACHE_DIR:
            return
        parent = super(_UserProvidedCacheLocator, cls)
        return parent.from_function(py_func, py_file)


class _UserWideCacheLocator(_SourceFileBackedLocatorMixin, _CacheLocator):
    """
    A locator for functions backed by a regular Python module or a
    frozen executable, cached into a user-wide cache directory.
    """

    def __init__(self, py_func, py_file):
        self._py_file = py_file
        self._lineno = py_func.__code__.co_firstlineno
        appdirs = AppDirs(appname="numba_cuda_mlir", appauthor=False)
        cache_dir = appdirs.user_cache_dir
        cache_subpath = self.get_suitable_cache_subpath(py_file)
        self._cache_path = os.path.join(cache_dir, cache_subpath)

    def get_cache_path(self):
        return self._cache_path

    @classmethod
    def from_function(cls, py_func, py_file):
        if not (os.path.exists(py_file) or getattr(sys, "frozen", False)):
            # Perhaps a placeholder (e.g. "<ipython-XXX>")
            # stop function exit if frozen, since it uses a temp placeholder
            return
        self = cls(py_func, py_file)
        try:
            self.ensure_cache_path()
        except OSError:
            # Cannot ensure the cache directory exists or is writable
            return
        return self


class _IPythonCacheLocator(_CacheLocator):
    """
    A locator for functions entered at the IPython prompt (notebook or other).
    """

    def __init__(self, py_func, py_file):
        self._py_file = py_file
        # Note IPython enhances the linecache module to be able to
        # inspect source code of functions defined on the interactive prompt.
        source = inspect.getsource(py_func)
        if isinstance(source, bytes):
            self._bytes_source = source
        else:
            self._bytes_source = source.encode("utf-8")

    def get_cache_path(self):
        # We could also use jupyter_core.paths.jupyter_runtime_dir()
        # In both cases this is a user-wide directory, so we need to
        # be careful when disambiguating if we don't want too many
        # conflicts (see below).
        try:
            from IPython.paths import get_ipython_cache_dir
        except ImportError:
            # older IPython version
            from IPython.utils.path import get_ipython_cache_dir
        return os.path.join(get_ipython_cache_dir(), "numba_cuda_mlir_cache")

    def get_source_stamp(self):
        return hashlib.sha256(self._bytes_source).hexdigest()

    def get_disambiguator(self):
        # Heuristic: we don't want too many variants being saved, but
        # we don't want similar named functions (e.g. "f") to compete
        # for the cache, so we hash the first two lines of the function
        # source (usually this will be the @jit decorator + the function
        # signature).
        firstlines = b"".join(self._bytes_source.splitlines(True)[:2])
        return hashlib.sha256(firstlines).hexdigest()[:10]

    @classmethod
    def from_function(cls, py_func, py_file):
        if not (
            py_file.startswith("<ipython-")
            or os.path.basename(os.path.dirname(py_file)).startswith("ipykernel_")
        ):
            return
        self = cls(py_func, py_file)
        try:
            self.ensure_cache_path()
        except OSError:
            # Cannot ensure the cache directory exists
            return
        return self


class _ZipCacheLocator(_SourceFileBackedLocatorMixin, _CacheLocator):
    """
    A locator for functions backed by Python modules within a zip archive.
    """

    def __init__(self, py_func, py_file):
        self._py_file = py_file
        self._lineno = py_func.__code__.co_firstlineno
        self._zip_path, self._internal_path = self._split_zip_path(py_file)
        # We use AppDirs at the moment. A more advanced version of this could also allow
        # a provided `cache_dir`, though that starts to create (cache location x source
        # type) number of cache classes.
        appdirs = AppDirs(appname="numba_cuda_mlir", appauthor=False)
        cache_dir = appdirs.user_cache_dir
        cache_subpath = self.get_suitable_cache_subpath(py_file)
        self._cache_path = os.path.join(cache_dir, cache_subpath)

    @staticmethod
    def _split_zip_path(py_file):
        path = Path(py_file)
        for i, part in enumerate(path.parts):
            if part.endswith(".zip"):
                zip_path = str(Path(*path.parts[: i + 1]))
                internal_path = str(Path(*path.parts[i + 1 :]))
                return zip_path, internal_path
        raise ValueError("No zip file found in path")

    def get_cache_path(self):
        return self._cache_path

    def get_source_stamp(self):
        st = os.stat(self._zip_path)
        return st.st_mtime, st.st_size

    @classmethod
    def from_function(cls, py_func, py_file):
        if ".zip" not in py_file:
            return None
        return cls(py_func, py_file)


class CacheImpl(metaclass=ABCMeta):
    """
    Provides the core machinery for caching.
    - implement how to serialize and deserialize the data in the cache.
    - control the filename of the cache.
    - provide the cache locator
    """

    _locator_classes = [
        _UserProvidedCacheLocator,
        _InTreeCacheLocator,
        _UserWideCacheLocator,
        _IPythonCacheLocator,
        _ZipCacheLocator,
    ]

    def __init__(self, py_func):
        self._lineno = py_func.__code__.co_firstlineno
        # Get qualname
        try:
            qualname = py_func.__qualname__
        except AttributeError:
            qualname = py_func.__name__
        # Find a locator
        source_path = inspect.getfile(py_func)
        for cls in self._locator_classes:
            locator = cls.from_function(py_func, source_path)
            if locator is not None:
                break
        else:
            raise RuntimeError(
                "cannot cache function %r: no locator available "
                "for file %r" % (qualname, source_path)
            )
        self._locator = locator
        # Use filename base name as module name to avoid conflict between
        # foo/__init__.py and foo/foo.py
        filename = inspect.getfile(py_func)
        modname = os.path.splitext(os.path.basename(filename))[0]
        fullname = "%s.%s" % (modname, qualname)
        abiflags = getattr(sys, "abiflags", "")
        self._filename_base = self.get_filename_base(fullname, abiflags)

    def get_filename_base(self, fullname, abiflags):
        # '<' and '>' can appear in the qualname (e.g. '<locals>') but
        # are forbidden in Windows filenames
        fixed_fullname = fullname.replace("<", "").replace(">", "")
        fmt = "%s-%s.py%d%d%s"
        return fmt % (
            fixed_fullname,
            self.locator.get_disambiguator(),
            sys.version_info[0],
            sys.version_info[1],
            abiflags,
        )

    @property
    def filename_base(self):
        return self._filename_base

    @property
    def locator(self):
        return self._locator

    @abstractmethod
    def reduce(self, data):
        "Returns the serialized form the data"
        pass

    @abstractmethod
    def rebuild(self, target_context, reduced_data):
        "Returns the de-serialized form of the *reduced_data*"
        pass

    @abstractmethod
    def check_cachable(self, data):
        "Returns True if the given data is cachable; otherwise, returns False."
        pass
